/* Langton's Ant
 * Author: Aaron Weeden, Shodor, 2015
 */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

// Declare constants
#define TIME_STEPS_PER_SIM 200
#define ANTS_PER_SIM 10
#define PX_PER_CELL_X 5
#define PX_PER_CELL_Y 5
#define PX_PER_ANT_X 2
#define PX_PER_ANT_Y 2
#define CELLS_PER_GRID_X 21
#define CELLS_PER_GRID_Y 21
#define RANDOM_SEED 0
#define PX_PER_GRID_X ((PX_PER_CELL_X) * (CELLS_PER_GRID_X))
#define PX_PER_GRID_Y ((PX_PER_CELL_Y) * (CELLS_PER_GRID_Y))

#define BLACK 0
#define WHITE 1

#define UP    0
#define LEFT  1
#define RIGHT 2
#define DOWN  3

void print(int const timeStep,
           int cellStates[CELLS_PER_GRID_X][CELLS_PER_GRID_Y],
           int const * antDirs, int const * antCellXs, int const * antCellYs) {
  printf("Time step %d\n", timeStep);
  for (int y = 0; y < CELLS_PER_GRID_Y; y++) {
    for (int x = 0; x < CELLS_PER_GRID_X; x++) {
      // Print cell color
      printf("%c", cellStates[x][y] == BLACK ? 'B' : '_');

      // Print ant direction
      bool isAntHere = false;
      for (int k = 0; k < ANTS_PER_SIM; k++) {
        if (antCellXs[k] == x && antCellYs[k] == y) {
          switch (antDirs[k]) {
            case UP:
              printf("U");
              break;
            case LEFT:
              printf("L");
              break;
            case RIGHT:
              printf("R");
              break;
            case DOWN:
              printf("D");
              break;
          }
          // Only print direction for one ant
          isAntHere = true;
          break;
        }
      }
      if (!isAntHere) {
        printf(" ");
      }
      printf(" ");
    }
    printf("\n");
  }
  printf("\n");
}

int main() {
  // Initialize cell states
  int cellStates[CELLS_PER_GRID_X][CELLS_PER_GRID_Y];
  for (int i = 0; i < CELLS_PER_GRID_X; i++) {
    for (int j = 0; j < CELLS_PER_GRID_Y; j++) {
      cellStates[i][j] = WHITE;
    }
  }

  // For each ant
  int antCellXs[ANTS_PER_SIM];
  int antCellYs[ANTS_PER_SIM];
  int antDirs[ANTS_PER_SIM];
  srandom(RANDOM_SEED);
  for (int i = 0; i < ANTS_PER_SIM; i++) {
    // Give the ant a position
    antCellXs[i] = random() % CELLS_PER_GRID_X;
    antCellYs[i] = random() % CELLS_PER_GRID_Y;

    // Give the ant a direction
    antDirs[i] = LEFT;
  }

  print(0, cellStates, antDirs, antCellXs, antCellYs);

  // Start the simulation
  for (int timeStep = 0; timeStep < TIME_STEPS_PER_SIM; timeStep++) {
    // Turn
    for (int i = 0; i < ANTS_PER_SIM; i++) {
      bool isBlack = (cellStates[antCellXs[i]][antCellYs[i]] == BLACK);
      if (     (antDirs[i] == UP    &&  isBlack) ||
               (antDirs[i] == DOWN  && !isBlack)) {
        antDirs[i] = LEFT;
      }
      else if ((antDirs[i] == UP    && !isBlack) ||
               (antDirs[i] == DOWN  &&  isBlack)) {
        antDirs[i] = RIGHT;
      }
      else if ((antDirs[i] == LEFT  &&  isBlack) ||
               (antDirs[i] == RIGHT && !isBlack)) {
        antDirs[i] = DOWN;
      }
      else if ((antDirs[i] == LEFT  && !isBlack) ||
               (antDirs[i] == RIGHT &&  isBlack)) {
        antDirs[i] = UP;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);

    // Update cells
    int nCs = 0;
    int cs[CELLS_PER_GRID_X * CELLS_PER_GRID_Y];
    for (int i = 0; i < ANTS_PER_SIM; i++) {
      int const x = antCellXs[i];
      int const y = antCellYs[i];
      bool isChecked = false;
      for (int j = 0; j < nCs; j++) {
        if (cs[j] == x * CELLS_PER_GRID_X + y) {
          isChecked = true;
          break;
        }
      }
      if (!isChecked) {
        if (cellStates[x][y] == BLACK) {
          cellStates[x][y] = WHITE;
        }
        else {
          cellStates[x][y] = BLACK;
        }
        // Mark this cell as checked
        cs[nCs++] = x * CELLS_PER_GRID_X + y;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);

    // Move
    for (int i = 0; i < ANTS_PER_SIM; i++) {
      switch (antDirs[i]) {
        case UP:
          if (antCellYs[i] == 0) {
            antCellYs[i] = CELLS_PER_GRID_Y-1;
          }
          else {
            antCellYs[i]--;
          }
          break;
        case LEFT:
          if (antCellXs[i] == 0) {
            antCellXs[i] = CELLS_PER_GRID_X-1;
          }
          else {
            antCellXs[i]--;
          }
          break;
        case RIGHT:
          if (antCellXs[i] == CELLS_PER_GRID_X-1) {
            antCellXs[i] = 0;
          }
          else {
            antCellXs[i]++;
          }
          break;
        case DOWN:
          if (antCellYs[i] == CELLS_PER_GRID_Y-1) {
            antCellYs[i] = 0;
          }
          else {
            antCellYs[i]++;
          }
          break;
      }
    }

    print(timeStep, cellStates, antDirs, antCellXs, antCellYs);
  } // Continue animation loop

  return 0;
}

